import {
    _mamoru_storage_delete,
    _mamoru_storage_get,
    _mamoru_storage_keys,
    _mamoru_storage_open_external,
    _mamoru_storage_set
} from "./imports";
import {Value} from "./value";

abstract class KvStorageRead {
    protected readonly connection_id: i32

    protected constructor(connection_id: i32) {
        this.connection_id = connection_id
    }

    /**
     * Gets the value of the key.
     * If the key is not found, returns null.
     */
    public get(key: string): Value | null {
        const buffer = _mamoru_storage_get(this.connection_id, key);

        if (buffer.byteLength === 0) {
            return null
        }

        return Value.fromBytes(buffer)
    }

    /**
     * Checks if the key exists.
     */
    public contains(key: string): bool {
        return this.get(key) !== null
    }

    /**
     * Returns keys with the given prefix.
     * If the prefix is empty, returns all keys.
     */
    public keys(prefix: string = ''): string[] {
        const buffer = _mamoru_storage_keys(this.connection_id, prefix);
        const values = Value.fromBytes(buffer).asList();

        let keys: string[] = [];

        for (let i = 0; i < values.length; i++) {
            keys.push(values[i].asString());
        }

        return keys
    }
}

abstract class KvStorageWrite extends KvStorageRead {
    protected constructor(connection_id: i32) {
        super(connection_id);
    }

    /**
     * Sets the value of the key.
     * If the key already exists, overwrites it and returns the old value.
     * Optionally, sets the time-to-live (TTL) in seconds.
     * If the TTL is zero, the key will not expire.
     */
    public set(key: string, value: Value, ttl: u32 = 0): Value | null {
        const buffer = _mamoru_storage_set(key, value.toBytes().buffer, ttl)

        if (buffer.byteLength === 0) {
            return null
        }

        return Value.fromBytes(buffer)
    }

    /**
     * Deletes the key.
     * If the key is not found, does nothing.
     */
    public delete(key: string): void {
        _mamoru_storage_delete(key)
    }
}

export class KvStorage extends KvStorageWrite {
    protected constructor(connection_id: i32) {
        super(connection_id);
    }

    /**
     * Open a KV storage for the current agent.
     */
    public static open(): KvStorage {
        return new KvStorage(-1)
    }

    /**
     * Opens a KV storage for the external agent id.
     * The storage is read-only.
     * If the external agent id is not found, current agent will fail.
     */
    public static open_external(external_agent_id: string): KvStorageRead {
        const connection_id = _mamoru_storage_open_external(external_agent_id)

        return new KvStorage(connection_id)
    }
}
