import {Decoder, Encoder} from "@wapc/as-msgpack/assembly";
import {i128, u128, u256} from "as-bignum/assembly";
import {Sizer, Writer} from "@wapc/as-msgpack";
import {msgPackReadUint8Array} from "./util";

const BOOL_VALUE_TAG: string = "b";
const U64_VALUE_TAG: string = "u64";
const U256_VALUE_TAG: string = "u256";
const STRING_VALUE_TAG: string = "s";
const LIST_VALUE_TAG: string = "l";
const STRUCT_VALUE_TAG: string = "st";
const BYTES_VALUE_TAG: string = "by";

export class Value {
    private readonly _type: string;

    protected constructor(type: string) {
        this._type = type;
    }

    public static fromBytes(data: ArrayBuffer): Value {
        let decoder = new Decoder(data);

        return Value.fromDecoder(decoder);
    }

    private static fromDecoder(decoder: Decoder): Value {
        const _ = decoder.readMapSize();
        const type = decoder.readString();

        if (type == BOOL_VALUE_TAG) {
            return new BoolValue(decoder.readBool());
        }

        if (type == U64_VALUE_TAG) {
            return new U64Value(decoder.readUInt64());
        }

        if (type == U256_VALUE_TAG) {
            const bytes = Uint8Array.wrap(decoder.readByteArray());
            const value = u256.fromUint8ArrayBE(bytes);

            return new U256Value(value);
        }

        if (type == STRING_VALUE_TAG) {
            return new StringValue(decoder.readString());
        }

        if (type == LIST_VALUE_TAG) {
            return new ListValue(decoder.readArray<Value>((decoder: Decoder) => {
                return Value.fromDecoder(decoder);
            }));
        }

        if (type == STRUCT_VALUE_TAG) {
            const _ = decoder.readArraySize();
            const ty = decoder.readString();

            return new StructValue(ty, decoder.readMap<string, Value>((decoder: Decoder) => {
                return decoder.readString();
            }, (decoder: Decoder) => {
                return Value.fromDecoder(decoder);
            }));
        }

        if (type == BYTES_VALUE_TAG) {
            let bytes = msgPackReadUint8Array(decoder);

            return new BytesValue(bytes);
        }

        throw new Error("unknown `Value` type: " + type);
    }

    public toBytes(): Uint8Array {
        const sizer = new Sizer();
        this.encode(sizer);

        const buffer = new ArrayBuffer(sizer.length);
        const encoder = new Encoder(buffer);

        this.encode(encoder);

        return Uint8Array.wrap(buffer);
    }

    protected encode(writer: Writer): void {
        writer.writeMapSize(2);
        writer.writeString(this._type);
        this.writeValue(writer);
    }

    protected writeValue(writer: Writer): void {
        // no-op for generic value
    }

    public isBool(): boolean {
        return this._type == BOOL_VALUE_TAG;
    }

    public asBool(): boolean {
        return (this as BoolValue).value;
    }

    public isU64(): boolean {
        return this._type == U64_VALUE_TAG;
    }

    public asU64(): u64 {
        return (this as U64Value).value;
    }

    public isString(): boolean {
        return this._type == STRING_VALUE_TAG;
    }

    public asString(): string {
        return (this as StringValue).value;
    }

    public asU128(): u128 {
        return u128.fromString(this.asString(), 16);
    }

    public asI128(): i128 {
        return i128.fromString(this.asString(), 16);
    }

    public isU256(): boolean {
        return this._type == U256_VALUE_TAG;
    }

    public asU256(): u256 {
        return (this as U256Value).value;
    }

    public isList(): boolean {
        return this._type == LIST_VALUE_TAG;
    }

    public asList(): Value[] {
        return (this as ListValue).values;
    }

    public isStruct(): boolean {
        return this._type == STRUCT_VALUE_TAG;
    }

    public asStruct(): StructValue {
        return (this as StructValue);
    }

    public isBytes(): boolean {
        return this._type == BYTES_VALUE_TAG;
    }

    public asBytes(): Uint8Array {
        return (this as BytesValue).value;
    }
}

export class BoolValue extends Value {
    private readonly _value: bool;

    public constructor(value: bool) {
        super(BOOL_VALUE_TAG);

        this._value = value;
    }

    public get value(): bool {
        return this._value;
    }

    protected writeValue(writer: Writer): void {
        writer.writeBool(this._value);
    }
}

export class U64Value extends Value {
    private readonly _value: u64;

    public constructor(value: u64) {
        super(U64_VALUE_TAG);

        this._value = value;
    }

    public get value(): u64 {
        return this._value;
    }

    protected writeValue(writer: Writer): void {
        writer.writeUInt64(this._value);
    }
}

export class U256Value extends Value {
    private readonly _value: u256;

    public constructor(value: u256) {
        super(U256_VALUE_TAG);

        this._value = value;
    }

    public get value(): u256 {
        return this._value;
    }

    protected writeValue(writer: Writer): void {
        writer.writeByteArray(this._value.toUint8Array(true).buffer);
    }
}

export class StringValue extends Value {
    private readonly _value: string;

    public constructor(value: string) {
        super(STRING_VALUE_TAG);

        this._value = value;
    }

    public get value(): string {
        return this._value;
    }

    protected writeValue(writer: Writer): void {
        writer.writeString(this._value);
    }
}

export class ListValue extends Value {
    private readonly _values: Value[];

    public constructor(value: Value[]) {
        super(LIST_VALUE_TAG);

        this._values = value;
    }

    public get values(): Value[] {
        return this._values;
    }

    protected writeValue(writer: Writer): void {
        writer.writeArray(this._values, (writer: Writer, value: Value) => {
            value.encode(writer);
        });
    }
}

export class StructValue extends Value {
    private readonly _ty: string;
    private readonly _value: Map<string, Value>;

    public constructor(ty: string, value: Map<string, Value>) {
        super(STRUCT_VALUE_TAG);

        this._ty = ty;
        this._value = value;
    }

    public get type(): string {
        return this._ty;
    }

    public get fields(): Map<string, Value> {
        return this._value;
    }

    protected writeValue(writer: Writer): void {
        writer.writeArraySize(2);
        writer.writeString(this._ty);

        writer.writeMap(this._value, (writer: Writer, key: string) => {
            writer.writeString(key);
        }, (writer: Writer, value: Value) => {
            value.encode(writer);
        });
    }
}

export class BytesValue extends Value {
    private readonly _value: Uint8Array;

    public constructor(value: Uint8Array) {
        super(BYTES_VALUE_TAG);

        this._value = value;
    }

    public get value(): Uint8Array {
        return this._value;
    }

    protected writeValue(writer: Writer): void {
        writer.writeArraySize(this.value.length);
        for (let i: i32 = 0; i < this.value.length; i++) {
            writer.writeUInt8(this.value[i]);
        }
    }
}
